#include <vector>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <memory>
#include <numeric>
#include <functional>
#include <sstream>
#include <limits>

#include "readcpt.h"

// read the distance, gradient fields in from the specified file
SignedDistanceRepresentation::SignedDistanceRepresentation(const std::string distanceFilename,
	const std::string gradientFilename)
{
	{
		{
			std::ifstream distanceFile( distanceFilename.c_str(), std::ios::binary );
			if( !distanceFile.good() )
			{
				std::ostringstream oss;
				oss << "Unable to open file '" << distanceFilename << "'.";
				throw DistanceFieldException( oss.str() );
			}

			Vec3 cartesianDomain[2];
			distanceFile.read( reinterpret_cast<char*>(cartesianDomain[0].Ref()), 3*sizeof(Real) );
			distanceFile.read( reinterpret_cast<char*>(cartesianDomain[1].Ref()), 3*sizeof(Real) );

			_bounds = BoundingBox3d( cartesianDomain[0], cartesianDomain[1] );
			distanceFile.read( reinterpret_cast<char*>(_gridDimensions), 3*sizeof(size_t) );
			_distance.resize( std::accumulate(_gridDimensions, _gridDimensions + 3, 1, std::multiplies<int>()) );
			distanceFile.read( reinterpret_cast<char*>(&_distance[0]),
				_distance.size()*sizeof( Real ) );
			unsigned int t = _distance.size();
			if( distanceFile.fail() )
				throw DistanceFieldException( "Error reading distance file" );
		}

		{
			std::ifstream gradientFile( gradientFilename.c_str(), std::ios::binary );
			if( !gradientFile.good() )
				throw DistanceFieldException( std::string("Unable to open file '") + gradientFilename +
					std::string( "'." ) );

			Real cartesianDomain[6];
			gradientFile.read( reinterpret_cast<char*>(cartesianDomain), 6*sizeof(Real) );

			int gridDimensions[3];
			gradientFile.read( reinterpret_cast<char*>(gridDimensions), 3*sizeof(size_t) );
			if( gridDimensions[0] != _gridDimensions[0]
				|| gridDimensions[1] != _gridDimensions[1]
				|| gridDimensions[2] != _gridDimensions[2] )
				throw DistanceFieldException( "Different grid dimensions between files!" );

			_gradientDistance.resize( std::accumulate(_gridDimensions, _gridDimensions + 3, 1, std::multiplies<int>()) );
			gradientFile.read( reinterpret_cast<char*>(&_gradientDistance[0]),
				_gradientDistance.size()*sizeof( Vec3 ) );
			if( gradientFile.fail() )
				throw DistanceFieldException( "Error reading gradient file" );
		}

		_delta = (_bounds.maximum() - _bounds.minimum());
		_delta[0] /= static_cast<Real>(_gridDimensions[0] - 1);
		_delta[1] /= static_cast<Real>(_gridDimensions[1] - 1);
		_delta[2] /= static_cast<Real>(_gridDimensions[2] - 1);
	}

	// initialize strides (stolen from the CPT code)
	_strides[0] = 1;
	for ( int i = 1; i < 3; ++i )
		_strides[i] = _strides[i-1] * _gridDimensions[i-1];

	std::cout << "Read distance file sampled at "
		<< _gridDimensions[0] << "x" << _gridDimensions[1] << "x" <<
		_gridDimensions[2] << std::endl;
	std::cout << "Cartesian domain is " << _bounds.minimum() << " -> " << std::endl
		<< "   " << _bounds.maximum() << std::endl;
	std::cout << "Min distance: " <<
		*std::min_element( _distance.begin(), _distance.end() ) << ", Max distance: " <<
		*std::max_element( _distance.begin(), _distance.end() ) << std::endl;

	std::cout << "Strides: " << _strides[0] << ", " << _strides[1] << ", " << _strides[2] << std::endl;
}

